@@ -35,8 +35,8 @@ MASK_LITERAL = 1
 MASK_EXPRESSION = 2
 MAXEXPANSION = 99
 
-FOLIAVERSION = '0.12.1'
-FQLVERSION = '0.2.4'
+FOLIAVERSION = '1.3.1'
+FQLVERSION = '0.3.1'
 
 class SyntaxError(Exception):
     pass
@@ -110,7 +110,7 @@ class UnparsedQuery(object):
                                 s2 = s[i+1:j]
                                 break
                     if not s2 is None:
-                        self.q.append(s2.replace('\\"','"')) #undo escaped quotes
+                        self.q.append(s2.replace('\\"','"').replace("\\n","\n")) #undo escaped quotes and newlines
                         self.mask.append(MASK_LITERAL)
                         i = j
                         begin = i+1
@@ -224,22 +224,26 @@ class Filter(object): #WHERE ....
                 operator = q[i+1]
                 if q[i] == "class":
                     v = lambda x,y='cls': getattr(x,y)
-                elif q[i] == "text":
-                    v = lambda x,y='text': getattr(x,'text')()
+                elif q[i] in ("text","value"):
+                    v = lambda x,y='text': getattr(x,'value') if isinstance(x, (folia.Description, folia.Comment, folia.Content)) else getattr(x,'text')()
                 else:
                     v = lambda x,y=q[i]: getattr(x,y)
+                if q[i] == 'confidence':
+                    cnv = float
+                else:
+                    cnv =  lambda x: x
                 if operator == '=' or operator == '==':
                     filters.append( lambda x,y=q[i+2],v=v : v(x) == y )
                 elif operator == '!=':
                     filters.append( lambda x,y=q[i+2],v=v : v(x) != y )
                 elif operator == '>':
-                    filters.append( lambda x,y=q[i+2],v=v : v(x) > y )
+                    filters.append( lambda x,y=cnv(q[i+2]),v=v : False if v(x) is None else v(x) > y )
                 elif operator == '<':
-                    filters.append( lambda x,y=q[i+2],v=v : v(x) < y )
+                    filters.append( lambda x,y=cnv(q[i+2]),v=v : False if v(x) is None else v(x) < y )
                 elif operator == '>=':
-                    filters.append( lambda x,y=q[i+2],v=v : v(x) >= y )
+                    filters.append( lambda x,y=cnv(q[i+2]),v=v : False if v(x) is None else v(x) >= y )
                 elif operator == '<=':
-                    filters.append( lambda x,y=q[i+2],v=v : v(x) <= y )
+                    filters.append( lambda x,y=cnv(q[i+2]),v=v : False if v(x) is None else v(x) <= y )
                 elif operator == 'CONTAINS':
                     filters.append( lambda x,y=q[i+2],v=v : v(x).find( y ) != -1 )
                 elif operator == 'NOTCONTAINS':
@@ -983,21 +987,27 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
         while i < l:
             if q.kw(i,'SUGGESTION'):
                 i+= 1
-                suggestion = ( {}, {} )
+                suggestion = ( {}, {} ) #subassignments, suggestionassignments
                 if isinstance(q[i], UnparsedQuery):
                     if not q[i].kw(0,'SUBSTITUTE') and not q[i].kw(0,'ADD'):
                         raise SyntaxError("Subexpression after SUGGESTION, expected ADD or SUBSTITUTE, got " + str(q[i]))
                     Correction.parsesubstitute(q[i],suggestion)
                     i += 1
-                elif q.kw(i,'MERGE'):
-                    suggestion[0]['merge'] = True
-                    i+= 1
-                elif q.kw(i,'SPLIT'):
-                    suggestion[0]['split'] = True
-                    i+= 1
-                elif q.kw(i,'DELETION'):
-                    #No need to do anything, DELETION is just to make things more explicit in the syntax, it will result in an empty suggestion
+                elif q.kw(i,'MERGE') or q.kw(i,'SPLIT'):
+                    if q.kw(i,'MERGE'):
+                        suggestion[1]['merge'] = True
+                    else:
+                        suggestion[1]['split'] = True
                     i+= 1
+                    if q.kw(i,'DELETION'): #No need to do anything, DELETION is just to make things more explicit in the syntax, it will result in an empty suggestion
+                        i+=1
+                    elif isinstance(q[i], UnparsedQuery):
+                        if not q[i].kw(0,'SUBSTITUTE') and not q[i].kw(0,'ADD'):
+                            raise SyntaxError("Subexpression after SUGGESTION, expected ADD or SUBSTITUTE, got " + str(q[i]))
+                        Correction.parsesubstitute(q[i],suggestion)
+                        i += 1
+                    elif not q.kw(i,'WITH'):
+                        i = getassignments(q, i, suggestion[0], focus) #subassignments (the actual element in the suggestion)
                 elif not q.kw(i,'WITH'):
                     i = getassignments(q, i, suggestion[0], focus) #subassignments (the actual element in the suggestion)
                 if q.kw(i,'WITH'):
@@ -1033,7 +1043,7 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
                     actionassignments['set'] = query.defaultsets[action.focus.Class.XMLTAG]
                 except KeyError:
                     actionassignments['set'] = query.doc.defaultset(action.focus.Class)
-            if folia.Attrib.ID in action.focus.Class.REQUIRED_ATTRIBS:
+            if action.focus.Class.REQUIRED_ATTRIBS and folia.Attrib.ID in action.focus.Class.REQUIRED_ATTRIBS:
                 actionassignments['id'] = getrandomid(query, "corrected." + action.focus.Class.XMLTAG + ".")
 
         kwargs = {}
@@ -1108,7 +1118,7 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
                         subassignments['set'] = query.doc.defaultset(action.focus.Class)
                 if focus and not self.bare: #copy all data within (we have to do this again for each suggestion as it will generate different ID suffixes)
                     inheritchildren = list(focus.copychildren(query.doc, True))
-                if folia.Attrib.ID in action.focus.Class.REQUIRED_ATTRIBS:
+                if action.focus.Class.REQUIRED_ATTRIBS and folia.Attrib.ID in action.focus.Class.REQUIRED_ATTRIBS:
                     subassignments['id'] = getrandomid(query, "suggestion.")
                 kwargs['suggestions'].append( folia.Suggestion(query.doc, action.focus.Class(query.doc, *inheritchildren,**subassignments), **suggestionassignments )   )
 
@@ -1168,8 +1178,10 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
 
             insertindex = 0
             #find insertion index:
+            if debug: print("[FQL EVALUATION DEBUG] Correction.insert - Finding insertion index for target ", repr(target), "  in ", repr(target.parent),file=sys.stderr)
             for i, e in enumerate(target.parent):
-                if e is target[0]:
+                if e is target:
+                    if debug: print("[FQL EVALUATION DEBUG] Correction.insert - Target ", repr(target), " found in ", repr(target.parent), " at index ", i,file=sys.stderr)
                     insertindex = i
                     break
 
@@ -1183,6 +1195,9 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
         kwargs['nooriginal'] = True #this is an insertion, there is no original
         kwargs = self.assemblesuggestions(query,content,debug,kwargs)
 
+        if debug: print("[FQL EVALUATION DEBUG] Correction.insert - Applying and returning correction ", repr(kwargs),file=sys.stderr)
+        return content['parent'].correct(**kwargs)
+
     def substitute(self, query, substitution, contextselector, debug):
         kwargs = {}
         if self.set:
@@ -1225,7 +1240,7 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
 
         kwargs = self.assemblesuggestions(query,substitution,debug,kwargs)
 
-        if debug: print("[FQL EVALUATION DEBUG] Correction.substitute - Returning correction",file=sys.stderr)
+        if debug: print("[FQL EVALUATION DEBUG] Correction.substitute - Applying and returning correction",file=sys.stderr)
         return substitution['parent'].correct(**kwargs)
 
 
@@ -1242,11 +1257,11 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
                         actionassignments['set'] = query.doc.defaultset(Class)
             actionassignments['id'] = "corrected.%08x" % random.getrandbits(32) #generate a random ID
             e = Class(query.doc, **actionassignments)
-            if debug: print("[FQL EVALUATION DEBUG] Correction.substitute - Adding to new",file=sys.stderr)
+            if debug: print("[FQL EVALUATION DEBUG] Correction.assemblesuggestions - Adding to new",file=sys.stderr)
             kwargs['new'].append(e)
             for subaction in subactions:
                 subaction.focus.autodeclare(query.doc)
-                if debug: print("[FQL EVALUATION DEBUG] Correction.substitute - Invoking subaction", subaction.action,file=sys.stderr)
+                if debug: print("[FQL EVALUATION DEBUG] Correction.assemblesuggestions - Invoking subaction", subaction.action,file=sys.stderr)
                 subaction(query, [e], debug ) #note: results of subactions will be silently discarded
 
         for subassignments, suggestionassignments in self.suggestions:
@@ -1273,15 +1288,25 @@ class Correction(object): #AS CORRECTION/SUGGESTION expression...
                         subassignments['set'] = query.doc.defaultset(action.focus.Class)
                 focus = action.focus
                 focus.autodeclare(query.doc)
-                if folia.Attrib.ID in focus.Class.REQUIRED_ATTRIBS:
+                if focus.Class.REQUIRED_ATTRIBS and folia.Attrib.ID in focus.Class.REQUIRED_ATTRIBS:
                     subassignments['id'] = getrandomid(query, "suggestion.")
                 suggestionchildren.append( focus.Class(query.doc, **subassignments))
                 action = action.nextaction
 
+            if debug: print("[FQL EVALUATION DEBUG] Correction.assemblesuggestions - Suggestionchildren: ", len(suggestionchildren),file=sys.stderr)
+
             if 'split' in suggestionassignments and suggestionassignments['split']:
-                suggestionassignments['split'] = focus.ancestor(folia.StructureElement).id
+                nextitem = substitution['parent'].next(substitution['parent'].__class__, None)
+                if nextitem:
+                    suggestionassignments['split'] = nextitem.id
+                else:
+                    del suggestionassignments['split']
             if 'merge' in suggestionassignments and suggestionassignments['merge']:
-                suggestionassignments['merge'] = focus.ancestor(folia.StructureElement).id
+                nextitem = substitution['parent'].next(substitution['parent'].__class__, None)
+                if nextitem:
+                    suggestionassignments['merge'] = nextitem.id
+                else:
+                    del suggestionassignments['merge']
             kwargs['suggestions'].append( folia.Suggestion(query.doc,*suggestionchildren, **suggestionassignments )   )
 
         return kwargs
@@ -1291,24 +1316,32 @@ def getassignments(q, i, assignments,  focus=None):
     l = len(q)
     while i < l:
         if q.kw(i, ('id','set','annotator','class','n')):
-            assignments[q[i]] = q[i+1]
+            if q[i+1] == 'NONE':
+                assignments[q[i]] = None
+            else:
+                assignments[q[i]] = q[i+1]
             i+=2
         elif q.kw(i,'confidence'):
-            try:
-                assignments[q[i]] = float(q[i+1])
-            except:
-                raise SyntaxError("Invalid value for confidence: " + str(q[i+1]))
+            if q[i+1] == 'NONE':
+                assignments[q[i]] = None
+            else:
+                try:
+                    assignments[q[i]] = float(q[i+1])
+                except:
+                    raise SyntaxError("Invalid value for confidence: " + str(q[i+1]))
             i+=2
         elif q.kw(i,'annotatortype'):
             if q[i+1] == "auto":
                 assignments[q[i]] = folia.AnnotatorType.AUTO
             elif q[i+1] == "manual":
                 assignments[q[i]] = folia.AnnotatorType.MANUAL
+            elif q[i+1] == "NONE":
+                assignments[q[i]] = None
             else:
                 raise SyntaxError("Invalid value for annotatortype: " + str(q[i+1]))
             i+=2
         elif q.kw(i,'text'):
-            if not focus is None and focus.Class is folia.TextContent:
+            if not focus is None and focus.Class in (folia.TextContent, folia.Description, folia.Comment):
                 key = 'value'
             else:
                 key = 'text'
@@ -1317,6 +1350,8 @@ def getassignments(q, i, assignments,  focus=None):
         elif q.kw(i, 'datetime'):
             if q[i+1] == "now":
                 assignments[q[i]] = datetime.datetime.now()
+            elif q[i+1] == "NONE":
+                assignments[q[i]] = None
             elif q[i+1].isdigit():
                 try:
                     assignments[q[i]] = datetime.datetime.fromtimestamp(q[i+1])
@@ -1531,8 +1566,12 @@ class Action(object): #Action expression
                                 if debug: print("[FQL EVALUATION DEBUG] Action - Applying EDIT to focus ", repr(focus),file=sys.stderr)
                                 for attr, value in action.assignments.items():
                                     if attr in ("text","value"):
-                                        if debug: print("[FQL EVALUATION DEBUG] Action - settext("+ value+ ") on focus ", repr(focus),file=sys.stderr)
-                                        focus.settext(value)
+                                        if isinstance(focus, (folia.Description, folia.Comment, folia.Content)):
+                                            if debug: print("[FQL EVALUATION DEBUG] Action - setting value ("+ value+ ") on focus ", repr(focus),file=sys.stderr)
+                                            focus.value = value
+                                        else:
+                                            if debug: print("[FQL EVALUATION DEBUG] Action - settext("+ value+ ") on focus ", repr(focus),file=sys.stderr)
+                                            focus.settext(value)
                                     elif attr == "class":
                                         if debug: print("[FQL EVALUATION DEBUG] Action - " + attr +  " = " + value + " on focus ", repr(focus),file=sys.stderr)
                                         focus.cls = value
@@ -1585,7 +1624,7 @@ class Action(object): #Action expression
 
                     isspan = issubclass(action.focus.Class, folia.AbstractSpanAnnotation)
 
-                    if not 'set' in action.assignments:
+                    if not 'set' in action.assignments and action.focus.Class not in (folia.Description, folia.Comment):
                         if action.focus.set and action.focus.set != "undefined":
                             action.assignments['set'] = action.focus.set
                         elif action.focus.Class.XMLTAG in query.defaultsets:
@@ -1593,6 +1632,7 @@ class Action(object): #Action expression
                         else:
                             action.assignments['set'] = action.focus.set = query.doc.defaultset(action.focus.Class)
 
+
                     if isinstance(contextselector, tuple) and len(contextselector) == 2:
                         targetselection = contextselector[0](*contextselector[1])
                     else:
@@ -1702,6 +1742,56 @@ class Context(object):
         self.defaultsets = {}
 
 class Query(object):
+    """This class represents an FQL query.
+    
+    Selecting a word with a particular text is done as follows, ``doc`` is an instance of :class:`pynlpl.formats.folia.Document`::
+
+        query = fql.Query('SELECT w WHERE text = "house"')
+        for word in query(doc):
+            print(word)  #this will be an instance of folia.Word
+
+    Regular expression matching can be done using the ``MATCHES`` operator::
+
+        query = fql.Query('SELECT w WHERE text MATCHES "^house.*$"')
+        for word in query(doc):
+            print(word)  
+
+    The classes of other annotation types can be easily queried as follows::
+
+        query = fql.Query('SELECT w WHERE :pos = "v"' AND :lemma = "be"')
+        for word in query(doc):
+            print(word) 
+
+    You can constrain your queries to a particular target selection using the ``FOR`` keyword::
+
+        query = fql.Query('SELECT w WHERE text MATCHES "^house.*$" FOR s WHERE text CONTAINS "sell"')
+        for word in query(doc):
+            print(word)
+
+    This construction also allows you to select the actual annotations. To select all people (a named entity) for words that are not John::
+
+        query = fql.Query('SELECT entity WHERE class = "person" FOR w WHERE text != "John"')
+        for entity in query(doc):
+            print(entity) #this will be an instance of folia.Entity
+
+    **FOR** statement may be chained, and Explicit IDs can be passed using the ``ID`` keyword::
+
+        query = fql.Query('SELECT entity WHERE class = "person" FOR w WHERE text != "John" FOR div ID "section.21"')
+        for entity in query(doc):
+            print(entity) 
+        
+    Sets are specified using the **OF** keyword, it can be omitted if there is only one for the annotation type, but will be required otherwise::
+
+        query = fql.Query('SELECT su OF "http://some/syntax/set" WHERE class = "np"')
+        for su in query(doc):
+            print(su) #this will be an instance of folia.SyntacticUnit
+
+    We have just covered just the **SELECT** keyword, FQL has other keywords for manipulating documents, such as **EDIT**, **ADD**, **APPEND** and **PREPEND**.
+
+    Note:
+        Consult the FQL documentation at https://github.com/proycon/foliadocserve/blob/master/README.rst for further documentation on the language.
+
+    """
     def __init__(self, q, context=Context()):
         self.action = None
         self.targets = None
